#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#define _MIT_POSIX_THREADS 1
#include <pthread.h>

#include "v3.h"
#include "cons.h"
#include "brain.h"

#include <errno.h>
#include <unistd.h>
/*static inline _syscall2(int,clone, int,flags,char *,stack)
*/
#ifdef NEED_POW2
static double pow2(double parm)
{
	return pow(2, parm);
}
#endif

static int Scales[V3_COLS];

static char Dots[8*32];
static char LogLine[V3_COLS+2];
static int DotCount=0;
static FILE *fp;

#define maxresult 1
#define maxkesto 15

static int Status=-1;
static int TataPos=0;

typedef struct
{
	float Results[maxresult];
	float Tata[maxkesto][V3_COLS];
} AnalyzeData;

static AnalyzeData ToTrain, NowAnalyze;

static void Init(void)
{
	int a;
	int Size[256];
	float Min, Max, MaxMinDiff;
	
	#define kelpaa(a) (a>=32 && a!=127) /* isprint(a) */
	
	fp = stdout;
	for(a=0; a<256; a++)
		if(kelpaa(a))
		{
			int x, y, c=0;
			unsigned char *b = FontData + a*FontHeight;
			for(y=0; y<FontHeight; y++)
			{
				for(x=0; x<8; x++)if(*b & (1<<x))c++;
				b++;
			}
			Size[a] = c;			
		}
		
	for(a=0; a<8*FontHeight; a++)
	{
		int b, c=255, B=' ';
		for(b=0; b<256; b++)
			if(kelpaa(b))
			{
				register int size = abs(Size[b] - a);
				if(size <= c)
					if(c!=size || (rand()&1))
					{
						c = size;
						B = b;
					}
			}
		Dots[DotCount++] = B;
	}
	
#if 0
	while(DotCount >= 2)
	{
		if(Dots[DotCount-1] != Dots[DotCount-2])break;
		DotCount--;
	}
#endif

#if 0
	for(a=0; a<DotCount; a++)
		fprintf(fp, "%c", Dots[a]);
	fprintf(fp, "\n");
#endif
		
	Min = log(V3_MIN) / log(2.0);
	Max = log(V3_MAX) / log(2.0);
	
	MaxMinDiff = Max-Min;

	for(a=0; a < V3_COLS; a++)
		Scales[a] = (int)pow2(a * MaxMinDiff / V3_COLS + Min)/2;
		
	#undef kelpaa
	
	InitCons();
	InitBrain(V3_COLS*maxkesto, V3_HIDDENLAYERCOUNT, maxresult);
	
	bzero(&NowAnalyze, sizeof(NowAnalyze));

	printf("\33(U");	
}

static void Done(void)
{
	if(fp && fp != stdout && fp != stderr)
	{
		fclose(fp);
		fp = NULL;
	}
	
	printf("\33(B");
	
	DoneCons();
	DoneBrain();	
}

static char *Attr = "\33[m";

static int BrainLock = 0;

static void QueueTrainer(int n)
{
	int a;
	while(BrainLock)usleep(10000);
	
	a=0;//	a = clone(0,0,0,CLONE_VM,NULL);
	
	printf("Clone gave %d\n", a);
	
	if(!a)
	{
		BrainLock=1;
		
		for(a=0; a<maxresult; a++)
			ToTrain.Results[a] = (a==n) ? 0.9 : 0.1;
			
		BrainTrain((float *)ToTrain.Tata, ToTrain.Results, 1E-6, 0.8);
		
		BrainLock=0;
		
		_exit(0);
	}
}

static void View(void)
{
	int a, ch, ednyt=Scales[0];
	
	#define MAXVOIMA 4
	float voimat[MAXVOIMA];
	int   mitka[MAXVOIMA];
	
	float best;
	
	if(!fp)return;
	
	memset(voimat, 0, sizeof voimat);
	
	for(a=0; a < V3_COLS; a++)
	{
		register int dy;
		register int nyt = Scales[a];
		register float temp = 0.0;
		
//		if(nyt>ednyt)ednyt++;
		
		for(dy=0, ch=ednyt; ch<=nyt; ch++)
		{
			dy++;
			temp += Power[ch]-Avg[ch]/AvgCount;
		}
		
		temp /= dy;		
		
		ednyt = nyt;
		
		temp = temp<0.0 ? 0.0 : sqrt(temp) / 28.0;
		
		for(ch=0; ch<MAXVOIMA; ch++)
			if(temp >= voimat[ch])
			{
				int b;
				
				for(b=ch; b<MAXVOIMA-1; b++)
				{
					voimat[b+1]=voimat[b];
					mitka [b+1]=mitka[b];
				}
				
				mitka[ch] = a;
				voimat[ch] = temp;
				break;
			}		
		
		for(dy=0; dy<maxkesto-1; dy++)
		{
			NowAnalyze.Tata[dy][a] = NowAnalyze.Tata[dy+1][a];
			if(Status >= 0)
				ToTrain.Tata[dy][a] = ToTrain.Tata[dy+1][a];
		}
		
		dy = (int)(temp * DotCount);
		
		if(dy >= DotCount)dy = DotCount-1;
		
		LogLine[a] = Dots[dy];
		
		if(Status >= 0)
			ToTrain.Tata[maxkesto-1][a] = temp;
		
		NowAnalyze.Tata[maxkesto-1][a] = temp;
	}		
	LogLine[a-1] = LogLine[0];
/*	LogLine[a] = 0;
*/	
	fprintf(fp, "%s|", Attr);
	for(a=0; a<V3_COLS; a++)
	{
		int b;
		for(b=0; b<MAXVOIMA; b++)if(mitka[b] == a){b=-1;break;}
		if(b<0)
			fprintf(fp, "%s%c%s", "\33[0;1;33m", LogLine[a], Attr);
		else
			fputc(LogLine[a], fp);
	}
	fprintf(fp, "|%d ", BrainLock);
	
	if(BrainLock)
		fprintf(fp, "* busy");
	else
	{
		UseBrain((float *)NowAnalyze.Tata, NowAnalyze.Results);
	
		best = 0.0;
		ch   = 0;
		
		fprintf(fp, "(");
	
		for(a=0; a<maxresult; a++)
		{
			if(a>0)fprintf(fp, " ");
			if(NowAnalyze.Results[a] > 0.3)fprintf(fp, "\33[1m");
			fprintf(fp, "%6.4f", NowAnalyze.Results[a]);
			if(NowAnalyze.Results[a] > 0.3)fprintf(fp, "\33[m");
			if(NowAnalyze.Results[a] > best)
			{
				best = NowAnalyze.Results[a];
				ch = a;
			}
		}
		
		fprintf(fp, ") -> '%d'", ch);
	}
	
	fprintf(fp, "\33[m\n");
	
	fflush(fp);
	
	if(Status >= 0)
		if(++TataPos >= maxkesto)
			goto StatusEnd;
	
	if(kbhit())
	{
		ch = getch();
		
		if(Status >= 0)
		{
StatusEnd:	Attr = "\33[m";
			Status = -1;
			
			memcpy(&ToTrain.Tata, &NowAnalyze.Tata,
				(maxkesto-TataPos) * V3_COLS * sizeof(float));
			
			QueueTrainer(Status);
		}
		else if(ch >= '0' && ch <= '9' && ch < '0'+maxresult)
		{
			Status = ch - '0';
			TataPos = 0;
			
			Attr = "\33[1;44m";
		}
		else if(ch == ' ')
		{
			memcpy(&ToTrain, &NowAnalyze, sizeof(ToTrain));
			
			QueueTrainer(-1);
		}
	}
}

Viewer V3 = {fb_not_required, 0, Init, Done, View};
